import cv2

try:
  from PyQt6.QtCore import pyqtSignal, Qt, QEventLoop, QPointF, QRectF, QSize, QSizeF, QTimer
  from PyQt6.QtGui import QColor, QFont, QImage, QPainter, QPixmap, QPolygonF, QTransform
  from PyQt6.QtWidgets import QApplication, QGridLayout, QLabel, QLayout, QHBoxLayout, QPushButton, QSlider, QSpinBox, QToolTip, QVBoxLayout, QWidget
  PYQT6 = True
except ImportError:
  from PyQt5.QtCore import pyqtSignal, Qt, QEventLoop, QPointF, QRectF, QSize, QSizeF, QTimer
  from PyQt5.QtGui import QColor, QFont, QImage, QPainter, QPixmap, QPolygonF, QTransform
  from PyQt5.QtWidgets import QApplication, QGridLayout, QLabel, QLayout, QHBoxLayout, QPushButton, QSlider, QSpinBox, QToolTip, QVBoxLayout, QWidget
  PYQT6 = False

import zebrazoom.videoFormatConversion.zzVideoReading as zzVideoReading


TITLE_FONT = QFont('Helvetica', 18, QFont.Weight.Bold, True)
LIGHT_YELLOW = '#FFFFE0'
LIGHT_CYAN = '#E0FFFF'
LIGHT_GREEN = '#90ee90'
GOLD = '#FFD700'
SPINBOX_STYLESHEET = '''
QSpinBox::down-button  {
  subcontrol-origin: border;
  subcontrol-position: center left;
  height: 20;
  width: 20;
}

QSpinBox::up-button  {
  subcontrol-origin: border;
  subcontrol-position: center right;
  height: 20;
  width: 20;
}'''


def apply_style(widget, **kwargs):
    font = kwargs.pop('font', None)
    if font is not None:
        widget.setFont(font)
    widget.setStyleSheet(';'.join('%s: %s' % (prop.replace('_', '-'), val)  for prop, val in kwargs.items()))
    return widget


def transformCoordinates(fromRect, toRect, point):
  transform = QTransform()
  QTransform.quadToQuad(QPolygonF((fromRect.topLeft(), fromRect.topRight(), fromRect.bottomLeft(), fromRect.bottomRight())),
                        QPolygonF((toRect.topLeft(), toRect.topRight(), toRect.bottomLeft(), toRect.bottomRight())), transform)
  return transform.map(point)


def _dialogClosed(loop, fn):
  def inner(*args, **kwargs):
    loop.exit()
    return fn(*args, **kwargs)
  return inner


def _getButtonsLayout(buttons, loop, dialog=None):
  buttonsLayout = QHBoxLayout()
  buttonsLayout.addStretch()

  def callback(cb):
    if cb is not None:
      cb()
    if dialog is not None:
      dialog.close()
    else:
      loop.exit()

  for text, *args in buttons:
    if len(args) == 1:
      cb, = args
      exitLoop = True
      enabledSignal = None
    elif len(args) == 2:
      cb, exitLoop = args
      enabledSignal = None
    else:
      assert len(args) == 3
      cb, exitLoop, enabledSignal = args
    button = QPushButton(text)
    if enabledSignal is not None:
      button.setEnabled(False)
      enabledSignal.connect(lambda enabled, btn=button: btn.setEnabled(enabled))
    if exitLoop:
      button.clicked.connect(lambda *args, cb=cb: callback(cb))
    else:
      button.clicked.connect(cb)
    buttonsLayout.addWidget(button, alignment=Qt.AlignmentFlag.AlignCenter)
  buttonsLayout.addStretch()
  return buttonsLayout


def showBlockingPage(layout, title=None, buttons=(), dialog=False, labelInfo=None, exitSignals=()):
  loop = QEventLoop()
  for signal in exitSignals:
    signal.connect(lambda *args: loop.exit())
  mainLayout = QVBoxLayout()
  if title is not None:
    mainLayout.addWidget(apply_style(QLabel(title), font=TITLE_FONT), alignment=Qt.AlignmentFlag.AlignCenter)
  mainLayout.addLayout(layout)
  mainLayout.addLayout(_getButtonsLayout(buttons, loop))
  app = QApplication.instance()
  assert app is not None
  temporaryPage = QWidget()
  temporaryPage.setLayout(mainLayout)
  stackedLayout = app.window.centralWidget().layout()
  stackedLayout.addWidget(temporaryPage)
  oldWidget = stackedLayout.currentWidget()
  with app.suppressBusyCursor():
    stackedLayout.setCurrentWidget(temporaryPage)
    if labelInfo is not None:
      img, label = labelInfo
      label.setMinimumSize(1, 1)
      label.show()
      setPixmapFromCv(img, label)
    loop.exec()
    stackedLayout.setCurrentWidget(oldWidget)
  stackedLayout.removeWidget(temporaryPage)


def showDialog(layout, title=None, buttons=(), dialog=False, labelInfo=None, timeout=None):
  dialog = QWidget()
  loop = QEventLoop()
  mainLayout = QVBoxLayout()
  mainLayout.addLayout(layout)
  mainLayout.addLayout(_getButtonsLayout(buttons, loop, dialog=dialog))
  app = QApplication.instance()
  if app is not None:
    app.registerWindow(dialog)
  dialog.setWindowTitle(title)
  dialog.move(0, 0)
  dialog.setWindowModality(Qt.WindowModality.ApplicationModal)
  dialog.setLayout(mainLayout)
  if labelInfo is not None:
    img, label = labelInfo
    height, width = img.shape[:2]
    label.setMinimumSize(width, height)
    layoutSize = mainLayout.totalSizeHint()
    label.setMinimumSize(1, 1)
    extraWidth = layoutSize.width() - width
    extraHeight = layoutSize.height() - height
  else:
    layoutSize = mainLayout.totalSizeHint()
  screenSize = QApplication.primaryScreen().availableSize()
  if layoutSize.width() > screenSize.width() or layoutSize.height() > screenSize.height():
    layoutSize.scale(screenSize, Qt.AspectRatioMode.KeepAspectRatio)
  dialog.setFixedSize(layoutSize)
  dialog.show()
  if labelInfo is not None:
    setPixmapFromCv(*labelInfo, QSize(layoutSize.width() - extraWidth, layoutSize.height() - extraHeight))
  dialog.closeEvent = _dialogClosed(loop, dialog.closeEvent)
  if timeout is not None:
    QTimer.singleShot(timeout, lambda: dialog.close())
    loop.exec()

def _cvToPixmap(img):
  if len(img.shape) == 2:
    height, width = img.shape
    fmt = QImage.Format.Format_Grayscale8
    bytesPerLine = width
  else:
    assert len(img.shape) == 3
    height, width, channels = img.shape
    fmt = QImage.Format.Format_BGR888
    bytesPerLine = width * channels
  return QPixmap.fromImage(QImage(img.data.tobytes(), width, height, bytesPerLine, fmt))


def setPixmapFromCv(img, label, preferredSize=None):
  if not label.isVisible():
    label.setPixmap(_cvToPixmap(img))
    return
  scaling = label.devicePixelRatio() if PYQT6 else label.devicePixelRatioF()
  if label.pixmap() is None or label.pixmap().isNull():
    label.hide()
    label.setPixmap(_cvToPixmap(img))
    label.show()
  if preferredSize is None:
    preferredSize = label.pixmap().size()
  size = preferredSize.scaled(label.size(), Qt.AspectRatioMode.KeepAspectRatio)
  img = cv2.resize(img, (int(size.width() * scaling), int(size.height() * scaling)))
  pixmap = _cvToPixmap(img)
  pixmap.setDevicePixelRatio(scaling)
  label.setPixmap(pixmap)


class SliderWithSpinbox(QWidget):
  valueChanged = pyqtSignal(int)

  def __init__(self, value, minimum, maximum, name=None):
    super().__init__()
    minimum = int(minimum)
    maximum = int(maximum)

    layout = QGridLayout()
    layout.setRowStretch(0, 1)
    layout.setColumnStretch(0, 1)
    layout.setRowStretch(3, 1)
    layout.setColumnStretch(5, 1)
    layout.setVerticalSpacing(0)

    minLabel = QLabel(str(minimum))
    layout.addWidget(minLabel, 1, 1, Qt.AlignmentFlag.AlignLeft)
    if name is not None:
      layout.addWidget(QLabel(name), 1, 2, Qt.AlignmentFlag.AlignCenter)
    maxLabel = QLabel(str(maximum))
    layout.addWidget(maxLabel, 1, 3, Qt.AlignmentFlag.AlignRight)
    slider = QSlider(Qt.Orientation.Horizontal)
    slider.setMinimumWidth(350)
    slider.setRange(minimum, maximum)
    slider.setValue(value)
    layout.addWidget(slider, 2, 1, 1, 3)

    spinbox = QSpinBox()
    spinbox.setStyleSheet(SPINBOX_STYLESHEET)
    spinbox.setMinimumWidth(90)
    spinbox.setRange(minimum, maximum)
    spinbox.setValue(value)
    layout.addWidget(spinbox, 2, 4)

    def spinboxValueChanged():
      value = spinbox.value()
      slider.setValue(value)
      self.valueChanged.emit(value)
    spinbox.valueChanged.connect(spinboxValueChanged)
    slider.valueChanged.connect(lambda: spinbox.setValue(slider.value()))

    layout.setContentsMargins(20, 5, 20, 5)
    self.setLayout(layout)

    self.value = spinbox.value
    self.minimum = spinbox.minimum
    self.maximum = spinbox.maximum
    self.setValue = lambda value: spinbox.setValue(value) or slider.setValue(value)
    self.setMinimum = lambda min_: spinbox.setMinimum(min_) or slider.setMinimum(min_) or minLabel.setText(str(int(min_)))
    self.setMaximum = lambda max_: spinbox.setMaximum(max_) or slider.setMaximum(max_) or maxLabel.setText(str(int(max_)))
    self.setRange = lambda min_, max_: spinbox.setRange(min_, max_) or slider.setRange(min_, max_)
    self.isSliderDown = slider.isSliderDown
    self.sliderWidth = slider.width
    self.setPosition = slider.setSliderPosition


def _chooseFrameLayout(cap, spinboxValues):
  layout = QVBoxLayout()

  video = QLabel()
  layout.addWidget(video, alignment=Qt.AlignmentFlag.AlignCenter)
  layout.setStretch(0, 1)

  sliderWithSpinbox = SliderWithSpinbox(*spinboxValues, name="Frame")

  def valueChanged():
    value = sliderWithSpinbox.value()
    cap.set(1, value)
    ret, frame = cap.read()
    setPixmapFromCv(frame, video)
  sliderWithSpinbox.valueChanged.connect(valueChanged)
  layout.addWidget(sliderWithSpinbox)

  return layout, video, sliderWithSpinbox

def chooseBeginningPage(app, videoPath, title, chooseFrameBtnText, chooseFrameBtnCb, extraButtonInfo=None):
  cap = zzVideoReading.VideoCapture(videoPath)
  cap.set(1, 1)
  ret, frame = cap.read()
  layout, label, valueWidget = _chooseFrameLayout(cap, (1, 0, cap.get(7) - 2))

  chooseEnd = True
  def back():
    nonlocal chooseEnd
    chooseEnd = None
  def trackWholeVideo():
    nonlocal chooseEnd
    chooseEnd = False
  buttonsLayout = QHBoxLayout()
  buttonsLayout.addStretch()
  backBtn = QPushButton("Back")
  backBtn.setObjectName("back")
  buttonsLayout.addWidget(backBtn)
  chooseFrameBtn = QPushButton(chooseFrameBtnText)
  def chooseFrameBtnClicked():
    app.configFile["firstFrame"] = valueWidget.value()
    chooseFrameBtnCb()
  chooseFrameBtn.clicked.connect(chooseFrameBtnClicked)
  buttonsLayout.addWidget(chooseFrameBtn)
  extraBtn = None
  if extraButtonInfo is not None:
    text, cb = extraButtonInfo
    extraBtn = QPushButton(text)
    extraBtn.clicked.connect(cb)
    buttonsLayout.addWidget(extraBtn)
  buttonsLayout.addStretch()
  layout.addLayout(buttonsLayout)
  page = QWidget()
  page.setLayout(layout)
  stackedLayout = app.window.centralWidget().layout()
  stackedLayout.addWidget(page)
  oldWidget = stackedLayout.currentWidget()
  with app.suppressBusyCursor():
    stackedLayout.setCurrentWidget(page)
    label.setMinimumSize(1, 1)
    label.show()
    setPixmapFromCv(frame, label)
  for btn in (backBtn, chooseFrameBtn) + () if extraBtn is None else (extraBtn,):
    btn.clicked.connect(lambda: stackedLayout.removeWidget(page))


def chooseEndPage(app, videoPath, title, chooseFrameBtnText, chooseFrameBtnCb):
  cap = zzVideoReading.VideoCapture(videoPath)
  maximum = cap.get(7) - 2
  cap.set(1, maximum)
  ret, frame = cap.read()
  layout, label, valueWidget = _chooseFrameLayout(cap, (maximum, app.configFile["firstFrame"] + 1, maximum))
  proceed = True
  def back():
    nonlocal proceed
    proceed = False
  buttonsLayout = QHBoxLayout()
  buttonsLayout.addStretch()
  backBtn = QPushButton("Back")
  backBtn.setObjectName("back")
  buttonsLayout.addWidget(backBtn)
  chooseFrameBtn = QPushButton(chooseFrameBtnText)
  def chooseFrameBtnClicked():
    app.configFile["lastFrame"] = valueWidget.value()
    chooseFrameBtnCb()
  chooseFrameBtn.clicked.connect(chooseFrameBtnClicked)
  buttonsLayout.addWidget(chooseFrameBtn)
  buttonsLayout.addStretch()
  layout.addLayout(buttonsLayout)
  page = QWidget()
  page.setLayout(layout)
  stackedLayout = app.window.centralWidget().layout()
  stackedLayout.addWidget(page)
  oldWidget = stackedLayout.currentWidget()
  with app.suppressBusyCursor():
    stackedLayout.setCurrentWidget(page)
    label.setMinimumSize(1, 1)
    label.show()
    setPixmapFromCv(frame, label)
  for btn in (backBtn, chooseFrameBtn):
    btn.clicked.connect(lambda: stackedLayout.removeWidget(page))


class _InteractiveLabelPoint(QLabel):
  pointSelected = pyqtSignal(bool)

  def __init__(self, width, height, selectingRegion):
    super().__init__()
    self._width = width
    self._height = height
    self._point = None
    self._selectingRegion = selectingRegion
    self._currentPosition = None
    self._tooltipShown = False
    if self._selectingRegion:
      self.setMouseTracking(True)

  def mouseMoveEvent(self, evt):
    if self._selectingRegion:
      self._currentPosition = evt.pos()
      self.update()

  def mousePressEvent(self, evt):
    self._point = evt.pos()
    self.update()
    self.pointSelected.emit(True)

  def mouseReleaseEvent(self, evt):
    if self._point is not None and not self._tooltipShown:
      QToolTip.showText(self.mapToGlobal(self._point), "If you aren't satisfied with the selection, click again.", self, self.rect(), 5000)
      self._tooltipShown = True

  def enterEvent(self, evt):
    QApplication.setOverrideCursor(Qt.CursorShape.CrossCursor)

  def leaveEvent(self, evt):
    QApplication.restoreOverrideCursor()
    self._currentPosition = None
    self.update()

  def paintEvent(self, evt):
    super().paintEvent(evt)
    if self._currentPosition is None and self._point is None:
      return
    qp = QPainter()
    qp.begin(self)
    if self._currentPosition is not None:
      qp.setPen(QColor(255, 0, 0))
      qp.drawLine(0, self._currentPosition.y(), self.width(), self._currentPosition.y())
      qp.drawLine(self._currentPosition.x(), 0, self._currentPosition.x(), self.height())
    if self._point is not None:
      qp.setBrush(QColor(255, 0, 0))
      qp.setPen(Qt.PenStyle.NoPen)
      qp.drawEllipse(self._point, 2, 2)
    qp.end()

  def resizeEvent(self, evt):
    super().resizeEvent(evt)
    self._size = self.size()

  def getCoordinates(self):
    if self._point is None:
      return 0, 0
    point = self._point
    if self._size.height() != self._height or self._size.width() != self._width:
      point = transformCoordinates(QRectF(QPointF(0, 0), QSizeF(self._size)), QRectF(QPointF(0, 0), QSizeF(self._width, self._height)), self._point)
    return point.x(), point.y()


def getPoint(frame, title, extraButtons=(), selectingRegion=False, backBtnCb=None, useNext=True):
  height, width = frame.shape[:2]

  layout = QVBoxLayout()

  video = _InteractiveLabelPoint(width, height, selectingRegion)
  extraButtons = tuple((text, lambda: cb(video), exitLoop) for text, cb, exitLoop in extraButtons)
  buttons = (("Back", backBtnCb, True),) if backBtnCb is not None else ()
  buttons += (("Next", None, True, video.pointSelected),) if useNext else ()
  layout.addWidget(video, alignment=Qt.AlignmentFlag.AlignCenter)
  if not useNext:
    video.pointSelected.connect(lambda: QApplication.restoreOverrideCursor())
  showBlockingPage(layout, title=title, buttons=buttons + extraButtons, labelInfo=(frame, video), exitSignals=() if useNext else (video.pointSelected,))
  return video.getCoordinates()


class _InteractiveLabelRect(QLabel):
  regionSelected = pyqtSignal(bool)

  def __init__(self, width, height):
    super().__init__()
    self._width = width
    self._height = height
    self._topLeft = None
    self._currentPosition = None
    self._bottomRight = None
    self._size = None
    self._tooltipShown = False
    self.setMouseTracking(True)

  def mousePressEvent(self, evt):
    if self._topLeft is None or self._bottomRight is not None:
      self._topLeft = evt.pos()
      self._bottomRight = None
      self._currentPosition = None
      self.regionSelected.emit(False)
    else:
      self._bottomRight = evt.pos()
      self.regionSelected.emit(True)
    self.update()

  def mouseReleaseEvent(self, evt):
    if self._bottomRight is not None and not self._tooltipShown:
      QToolTip.showText(self.mapToGlobal(self._bottomRight), "If you aren't satisfied with the selection, click again.", self, self.rect(), 5000)
      self._tooltipShown = True

  def mouseMoveEvent(self, evt):
    self._currentPosition = evt.pos()
    self.update()

  def enterEvent(self, evt):
    QApplication.setOverrideCursor(Qt.CursorShape.CrossCursor)

  def leaveEvent(self, evt):
    QApplication.restoreOverrideCursor()
    self._currentPosition = None
    self.update()

  def paintEvent(self, evt):
    super().paintEvent(evt)
    if self._currentPosition is None and self._topLeft is None:
      return
    qp = QPainter()
    qp.begin(self)
    if self._currentPosition is not None and \
        (self._topLeft is None and self._bottomRight is None or self._bottomRight is not None):
      qp.setPen(QColor(255, 0, 0))
      qp.drawLine(0, self._currentPosition.y(), self.width(), self._currentPosition.y())
      qp.drawLine(self._currentPosition.x(), 0, self._currentPosition.x(), self.height())
    if self._topLeft is not None:
      qp.setPen(QColor(0, 0, 255))
      x = self._topLeft.x()
      y = self._topLeft.y()
      bottomRight = self._bottomRight or self._currentPosition
      if bottomRight is None:
        qp.drawPoint(x, y)
      else:
        width = bottomRight.x() - x
        height = bottomRight.y() - y
        qp.drawRect(x, y, width, height)
    qp.end()

  def resizeEvent(self, evt):
    super().resizeEvent(evt)
    self._size = self.size()

  def getCoordinates(self):
    if self._topLeft is None or self._bottomRight is None:
      return [0, 0], [0, 0]
    points = (self._topLeft, self._bottomRight)
    if self._size.height() != self._height or self._size.width() != self._width:
      points = (transformCoordinates(QRectF(QPointF(0, 0), QSizeF(self._size)), QRectF(QPointF(0, 0), QSizeF(self._width, self._height)), point) for point in points)
    return ([point.x(), point.y()] for point in points)


def getRectangle(frame, title, backBtnCb=None):
  height, width, _ = frame.shape

  layout = QVBoxLayout()

  video = _InteractiveLabelRect(width, height)
  layout.addWidget(video, alignment=Qt.AlignmentFlag.AlignCenter)
  if backBtnCb is not None:
    buttons = (("Back", backBtnCb, True), ("Next", None, True, video.regionSelected))
  else:
    buttons = (("Next", None, True, video.regionSelected),)
  showBlockingPage(layout, title=title, buttons=buttons, labelInfo=(frame, video))

  return video.getCoordinates()


def addToHistory(fn):
  def inner(*args, **kwargs):
    app = QApplication.instance()
    configFileState = app.configFile.copy()
    def restoreState():
      app.configFile.clear()
      app.configFile.update(configFileState)
      del app.configFileHistory[-1:]
      fn(*args, **kwargs)
    app.configFileHistory.append(restoreState)
    return fn(*args, **kwargs)
  return inner



from zebrazoom.dataAnalysis.dataanalysis.reapplyClustering import reapplyClustering
import matplotlib.pyplot as plt
import numpy as np
import pickle

def launchOptimalClusterNumberSearch():

  outputFolderResult = 'zebrazoom/dataAnalysis/resultsClustering/parkinsonFreelySwimForClustering_OutliersRemoved_BASSstyleParam2/activeLearning/'
  nbConditions    = 2
  nbClusterToTest = 7

  KMeans = {'silhouetteScore':       np.array([-1. for i in range(0, 2 + nbClusterToTest)]),
            'calinskiHarabaszScore': np.array([-1. for i in range(0, 2 + nbClusterToTest)]),
            'daviesBouldinScore':    np.array([-1. for i in range(0, 2 + nbClusterToTest)]),
            'elbowMethod':           np.array([-1. for i in range(0, 2 + nbClusterToTest)])}

  GaussianMixture = {'silhouetteScore':       np.array([-1. for i in range(0, 2 + nbClusterToTest)]),
                     'calinskiHarabaszScore': np.array([-1. for i in range(0, 2 + nbClusterToTest)]),
                     'daviesBouldinScore':    np.array([-1. for i in range(0, 2 + nbClusterToTest)]),
                     'elbowMethod':           np.array([-1. for i in range(0, 2 + nbClusterToTest)])}

  modelUsed = 'KMeans'
  for nbCluster in range(2, nbClusterToTest + 2):
    [silhouetteScore, calinskiHarabaszScore, daviesBouldinScore, elbowMethod] = reapplyClustering(modelUsed, nbConditions, nbCluster, outputFolderResult)
    KMeans['silhouetteScore'][nbCluster]       = silhouetteScore
    KMeans['calinskiHarabaszScore'][nbCluster] = calinskiHarabaszScore
    KMeans['daviesBouldinScore'][nbCluster]    = daviesBouldinScore
    KMeans['elbowMethod'][nbCluster]           = elbowMethod

  modelUsed = 'GaussianMixture'
  for nbCluster in range(2, nbClusterToTest + 2):
    [silhouetteScore, calinskiHarabaszScore, daviesBouldinScore, elbowMethod] = reapplyClustering(modelUsed, nbConditions, nbCluster, outputFolderResult)
    GaussianMixture['silhouetteScore'][nbCluster]       = silhouetteScore
    GaussianMixture['calinskiHarabaszScore'][nbCluster] = calinskiHarabaszScore
    GaussianMixture['daviesBouldinScore'][nbCluster]    = daviesBouldinScore
    GaussianMixture['elbowMethod'][nbCluster]           = elbowMethod

  with open('scoresComparaison.pkl', 'wb') as handle:
    pickle.dump([KMeans, GaussianMixture], handle, protocol=pickle.HIGHEST_PROTOCOL)

  # with open('scoresComparaison.pkl', 'rb') as handle:
    # b = pickle.load(handle)
  # KMeans = b[0]
  # GaussianMixture = b[1]

  fig, tabAx = plt.subplots(2, 2, figsize=(22.9, 8.8))

  tabAx[0, 0].plot([i for i in range(2, nbClusterToTest + 2)], KMeans['silhouetteScore'][2:nbClusterToTest + 2], label='KMeans')
  tabAx[0, 1].plot([i for i in range(2, nbClusterToTest + 2)], KMeans['calinskiHarabaszScore'][2:nbClusterToTest + 2], label='KMeans')
  tabAx[1, 0].plot([i for i in range(2, nbClusterToTest + 2)], KMeans['daviesBouldinScore'][2:nbClusterToTest + 2], label='KMeans')
  tabAx[1, 1].plot([i for i in range(2, nbClusterToTest + 2)], KMeans['elbowMethod'][2:nbClusterToTest + 2], label='KMeans')

  tabAx[0, 0].plot([i for i in range(2, nbClusterToTest + 2)], GaussianMixture['silhouetteScore'][2:nbClusterToTest + 2], label='GaussianMixture')
  tabAx[0, 1].plot([i for i in range(2, nbClusterToTest + 2)], GaussianMixture['calinskiHarabaszScore'][2:nbClusterToTest + 2], label='GaussianMixture')
  tabAx[1, 0].plot([i for i in range(2, nbClusterToTest + 2)], GaussianMixture['daviesBouldinScore'][2:nbClusterToTest + 2], label='GaussianMixture')
  # tabAx[1, 1].plot([i for i in range(2, nbClusterToTest + 2)], GaussianMixture['elbowMethod'][2:nbClusterToTest + 2], label='GaussianMixture')
  
  tabAx[0, 0].set_title("Silhouette Score")
  tabAx[0, 1].set_title("Calinski Harabasz Score")
  tabAx[1, 0].set_title("Davies Bouldin Score")
  tabAx[1, 1].set_title("Elbow Method")
  
  plt.legend()
  
  plt.savefig('scoresComparaison.png')



import sys
import os
import json
import scipy.io as io
from scipy.interpolate import UnivariateSpline
import scipy.interpolate as interp
from scipy.io import savemat
import numpy as np
from scipy.signal import find_peaks
import pandas as pd
import math
import zebrazoom.code.popUpAlgoFollow as popUpAlgoFollow

def IsMinOrMax(maxpeaks, minpeaks, val ):
  is_minOrMax = 0
  r = len(maxpeaks)
  for i in range(0,r):
    if maxpeaks[i] == val:
      is_minOrMax=1
  r = len(minpeaks)
  for i in range(0,r):
    if minpeaks[i] == val:
      is_minOrMax=1
  return is_minOrMax

# def calculateAllTailAngles(curbout, tailAngleSmoothingFactor, hyperparameters):
  # nbFramesTakenIntoAccount = len(curbout["TailAngle_Raw"])
  # fin = curbout["BoutEnd"] - curbout["BoutStart"] + 1
  # headx = curbout["HeadX"]
  # heady = curbout["HeadY"]
  # tailx = curbout["TailX_VideoReferential"]
  # taily = curbout["TailY_VideoReferential"]
  # heading = curbout["Heading"]
  # tailangles_arr = np.zeros((nbFramesTakenIntoAccount,8))
  
  # for i in range(min(len(headx),tailangles_arr.shape[0])):
    # if len(taily[i]) > 3 and len(tailx[i]) > 3:
      # ang = np.arctan2(heady[i] - taily[i][-3], headx[i] - tailx[i][-3])
      # for j in range(1, tailangles_arr.shape[1]):
        # ang2 = np.arctan2(heady[i] - taily[i][j], headx[i] - tailx[i][j])
        # delang = ang2 - ang
        # if np.abs(delang) < np.pi:
          # tailangles_arr[i,j] = delang
        # elif delang > np.pi:
          # tailangles_arr[i,j] = delang - 2*np.pi
        # elif delang < -np.pi:
          # tailangles_arr[i,j] = 2*np.pi + delang
    # else:
      # for j in range(tailangles_arr.shape[1]):
        # tailangles_arr[i,j] = 0
  # tailangles_arr = np.transpose(tailangles_arr)
  # tailangles_arr = np.append(tailangles_arr, np.array([[-val for val in curbout["TailAngle_Raw"]]]), axis=0)
  
  # tailangles_arr_smoothed = np.zeros((0, nbFramesTakenIntoAccount))
  # for angle_raw in tailangles_arr:
    # rolling_window = hyperparameters["tailAngleMedianFilter"]
    # if rolling_window > 0:
      # shift = int(-rolling_window / 2)
      # angle_median = np.array(pd.Series(angle_raw).rolling(rolling_window).median())
      # angle_median = np.roll(angle_median, shift)
      # for ii in range(0, rolling_window):
        # angle_median[ii] = angle_raw[ii]
      # for ii in range(len(angle_median)-rolling_window,len(angle_median)):
        # angle_median[ii] = angle_raw[ii]
    # else:
      # angle_median = angle_raw
    # tailToSmooth = angle_median
    # x = np.linspace(0, 1, len(tailToSmooth))
    # s = UnivariateSpline(x, tailToSmooth, s=tailAngleSmoothingFactor)
    # tailSmoothed     = s(x)
    # tailSmoothed2    = np.zeros((1, nbFramesTakenIntoAccount))
    # tailSmoothed2[0] = tailSmoothed
    # tailangles_arr_smoothed = np.append(tailangles_arr_smoothed, tailSmoothed2, axis=0)
  # return [tailangles_arr, tailangles_arr_smoothed]


def createSuperStruct(dataPerWell, wellPositions, hyperparameters):

  if (hyperparameters["freqAlgoPosFollow"] != 0):
    print("Starting the creation of the super structure")

  nbWells                      = hyperparameters["nbWells"]
  wellsAreRectangles           = hyperparameters["wellsAreRectangles"]
  tailAngleSmoothingFactor     = hyperparameters["tailAngleSmoothingFactor"]
  windowForLocalBendMinMaxFind = hyperparameters["windowForLocalBendMinMaxFind"]

  videoDataResults  = {}
  wellPoissMouv     = []
  
  for numWell in range(0,nbWells):
    
    item = {}
    
    j = dataPerWell[numWell]
    
    nbMouv = len(j)
    
    tab  = [[] for idAnimal in range(0, hyperparameters["nbAnimalsPerWell"])]
    
    for i in range(0, nbMouv):
    
      item = j[i]
      
      if hyperparameters["trackTail"]:
        
        angle_raw = item["TailAngle_Raw"]
        
        if len(angle_raw) > 10:
        
          rolling_window = hyperparameters["tailAngleMedianFilter"]
          if rolling_window > 0:
            shift = int(-rolling_window / 2)
            angle_median = np.array(pd.Series(angle_raw).rolling(rolling_window).median())
            angle_median = np.roll(angle_median, shift)
            for ii in range(0, rolling_window):
              angle_median[ii] = angle_raw[ii]
            for ii in range(len(angle_median)-rolling_window,len(angle_median)):
              angle_median[ii] = angle_raw[ii]
          else:
            angle_median = angle_raw
        
          x = np.linspace(0, 1, len(angle_median))
          
          # if True: # Original method
          s = UnivariateSpline(x, angle_median, s=tailAngleSmoothingFactor) # This might be non-terminating sometimes, need to investigate this in the future
          TailAngle_smoothed = s(x) 
          
          item['TailAngle_smoothed'] = TailAngle_smoothed.tolist()
        else:
          TailAngle_smoothed = np.array(angle_raw)
          item['TailAngle_smoothed'] = angle_raw
        
        # if hyperparameters["calculateAllTailAngles"]:
          # [tailangles_arr, tailangles_arr_smoothed] = calculateAllTailAngles(item, tailAngleSmoothingFactor, hyperparameters)
          # item["allTailAngles"]         = tailangles_arr.tolist()
          # item["allTailAnglesSmoothed"] = tailangles_arr_smoothed.tolist()
        
        if hyperparameters['extractAdvanceZebraParameters']:
        
          maxDiffPeakToPeak = 0
          maxAngle = max([TailAngle_smoothed[l] for l in range(0, len(TailAngle_smoothed))])
          minAngle = min([TailAngle_smoothed[l] for l in range(0, len(TailAngle_smoothed))])
          maxDiffPeakToPeak = maxAngle - minAngle
          
          minProminenceForBendsDetect = hyperparameters["minProminenceForBendsDetect"]
          if minProminenceForBendsDetect == -1:
            minProminenceForBendsDetect = maxDiffPeakToPeak / 10
          
          maxpeaks, properties = find_peaks(TailAngle_smoothed, prominence=minProminenceForBendsDetect, width=hyperparameters["windowForLocalBendMinMaxFind"])
          minpeaks, properties = find_peaks(-TailAngle_smoothed, prominence=minProminenceForBendsDetect, width=hyperparameters["windowForLocalBendMinMaxFind"])
          
          if (len(minpeaks) + len(maxpeaks) >= hyperparameters["minNbPeaksForBoutDetect"] ):
          
            ind = 1
            minDiffBetweenSubsequentBendAmp = hyperparameters["minDiffBetweenSubsequentBendAmp"]
            
            Bend_Timing = []
            lastTailValue = 100000

            for i in range(item['BoutStart'],item['BoutEnd']+1):
            
              minOrMax = IsMinOrMax(maxpeaks, minpeaks, i - item['BoutStart'] )
              
              if minOrMax:
                if (abs(lastTailValue - TailAngle_smoothed[i - item['BoutStart'] ]) > minDiffBetweenSubsequentBendAmp):
                  Bend_Timing.append(i - item['BoutStart'] + 1) 
                  lastTailValue = TailAngle_smoothed[i- item['BoutStart'] ]
              
              if (len(Bend_Timing) == 1) and (abs(lastTailValue) < hyperparameters["minFirstBendValue"]):
                Bend_Timing = []
                
            if len(Bend_Timing) > 3:
              Bend_Timing2 = []
              if hyperparameters["doubleCheckBendMinMaxStatus"]:
                # Checking first bend
                bendId = 0
                ind2 = Bend_Timing[bendId]
                ind3 = Bend_Timing[bendId+1]
                cond1 = (TailAngle_smoothed[0] < TailAngle_smoothed[ind2-1]) and (TailAngle_smoothed[ind3-1] < TailAngle_smoothed[ind2-1])
                cond2 = (TailAngle_smoothed[0] > TailAngle_smoothed[ind2-1]) and (TailAngle_smoothed[ind3-1] > TailAngle_smoothed[ind2-1])
                if cond1 or cond2:
                  Bend_Timing2.append(ind2)
                # Checking bends in the middle
                for bendId in range(1, len(Bend_Timing)-1):
                  ind1 = Bend_Timing[bendId-1]
                  ind2 = Bend_Timing[bendId]
                  ind3 = Bend_Timing[bendId+1]
                  cond1 = (TailAngle_smoothed[ind1-1] < TailAngle_smoothed[ind2-1]) and (TailAngle_smoothed[ind3-1] < TailAngle_smoothed[ind2-1])
                  cond2 = (TailAngle_smoothed[ind1-1] > TailAngle_smoothed[ind2-1]) and (TailAngle_smoothed[ind3-1] > TailAngle_smoothed[ind2-1])
                  if cond1 or cond2:
                    Bend_Timing2.append(ind2)
                # Checking last bend
                bendId = len(Bend_Timing)-1
                ind1 = Bend_Timing[bendId-1]
                ind2 = Bend_Timing[bendId]
                n = len(TailAngle_smoothed)-1
                cond1 = (TailAngle_smoothed[ind1-1] < TailAngle_smoothed[ind2-1]) and (TailAngle_smoothed[n] < TailAngle_smoothed[ind2-1])
                cond2 = (TailAngle_smoothed[ind1-1] > TailAngle_smoothed[ind2-1]) and (TailAngle_smoothed[n] > TailAngle_smoothed[ind2-1])
                if cond1 or cond2:
                  Bend_Timing2.append(ind2)
                Bend_Timing = Bend_Timing2
            
            if hyperparameters["removeFirstSmallBend"] and len(Bend_Timing) > 1:
              ind1 = Bend_Timing[0]
              ind2 = Bend_Timing[1]
              if abs(TailAngle_smoothed[ind1-1]) < abs(TailAngle_smoothed[ind2-1]) / hyperparameters["removeFirstSmallBend"]:
                Bend_Timing.pop(0)
            
            item['Bend_Timing'] = Bend_Timing
            
            Bend_TimingAbsolute = [val + item['BoutStart'] for val in Bend_Timing]
            
            item['Bend_TimingAbsolute'] = Bend_TimingAbsolute
            
            Bend_Amplitude = [TailAngle_smoothed[val-1] for val in Bend_Timing]
            item['Bend_Amplitude'] = Bend_Amplitude
            
            tab[item["AnimalNumber"]].append(item)
            
          else:
            item['Bend_Timing'] = []
            item['Bend_TimingAbsolute'] = []
            item['Bend_Amplitude'] = []
            item['TailAngle_raw'] = []
            item['TailAngle_smoothed'] = []
            tab[item["AnimalNumber"]].append(item)
          
        else:
          item['TailAngle_raw'] = []
          item['TailAngle_smoothed'] = []
          tab[item["AnimalNumber"]].append(item)
      else:
        tab[item["AnimalNumber"]].append(item)
      
    wellPoissMouv.append(tab)
  
  videoDataResults['wellPoissMouv'] = wellPoissMouv
  videoDataResults['wellPositions'] = wellPositions
  videoDataResults['firstFrame']    = hyperparameters["firstFrame"]
  videoDataResults['lastFrame']    = hyperparameters["lastFrame"]
  if hyperparameters["videoFPS"]:
    videoDataResults['fps']    = hyperparameters["videoFPS"]
    videoDataResults['videoFPS']    = hyperparameters["videoFPS"]
  if hyperparameters["videoPixelSize"]:
    videoDataResults['videoPixelSize']    = hyperparameters["videoPixelSize"]
  
  path = os.path.join(os.path.join(hyperparameters["outputFolder"], hyperparameters["videoName"]), 'results_' + hyperparameters["videoName"] + '.txt')
  
  print("createSuperStruct:", path)
  
  with open(path, 'w') as outfile:
    json.dump(videoDataResults, outfile)
    
  if hyperparameters["saveSuperStructToMatlab"]:
    matlabPath = os.path.join(os.path.join(hyperparameters["outputFolder"], hyperparameters["videoName"]), 'results_' + hyperparameters["videoName"] + '.mat')
    videoDataResults2 = {}
    videoDataResults2['videoDataResults'] = videoDataResults
    savemat(matlabPath, videoDataResults2)
    
  if (hyperparameters["freqAlgoPosFollow"] != 0):
    print("Super Structure created")
  if hyperparameters["popUpAlgoFollow"]:
    popUpAlgoFollow.prepend("Super Structure created")

  return videoDataResults



import webbrowser

try:
  from PyQt6.QtCore import Qt
  from PyQt6.QtGui import QFont, QIntValidator
  from PyQt6.QtWidgets import QLabel, QWidget, QPushButton, QLineEdit, QCheckBox, QVBoxLayout, QRadioButton, QButtonGroup
except ImportError:
  from PyQt5.QtCore import Qt
  from PyQt5.QtGui import QFont, QIntValidator
  from PyQt5.QtWidgets import QLabel, QWidget, QPushButton, QLineEdit, QCheckBox, QVBoxLayout, QRadioButton, QButtonGroup

import zebrazoom.code.util as util


class ChooseVideoToTroubleshootSplitVideo(QWidget):
  def __init__(self, controller):
    super().__init__(controller.window)
    self.controller = controller
    self.preferredSize = (750, 500)

    layout = QVBoxLayout()
    layout.addWidget(util.apply_style(QLabel("Troubleshooting.", self), font=controller.title_font), alignment=Qt.AlignmentFlag.AlignCenter)
    linkBtn = util.apply_style(QPushButton("First View Tracking Troubleshooting Tips", self), background_color=util.GOLD)
    linkBtn.clicked.connect(lambda: webbrowser.open_new("https://github.com/oliviermirat/ZebraZoom/blob/master/TrackingTroubleshooting.md"))
    layout.addWidget(linkBtn, alignment=Qt.AlignmentFlag.AlignCenter)
    layout.addWidget(QLabel("If the previous tracking troubleshooting tips where not enough to solve the issue, you can create a smaller sub-video to send to ZebraZoom's developers for troubleshooting.", self), alignment=Qt.AlignmentFlag.AlignCenter)

    selectVideoBtn = QPushButton("Select the video to troubleshoot.", self)
    selectVideoBtn.clicked.connect(lambda: controller.chooseVideoToTroubleshootSplitVideo(controller))
    layout.addWidget(selectVideoBtn, alignment=Qt.AlignmentFlag.AlignCenter)
    layout.addWidget(util.apply_style(QLabel("If you are having issues running the tracking on a video or creating a good configuration file for a video", self), font=QFont("Helvetica", 10)), alignment=Qt.AlignmentFlag.AlignCenter)
    layout.addWidget(util.apply_style(QLabel("you can create a sub-video centered around a bout of movement and send this smaller sub-video to info@zebrazoom.org in order for us to help troubleshoot.", self), font=QFont("Helvetica", 10)), alignment=Qt.AlignmentFlag.AlignCenter)
    layout.addWidget(util.apply_style(QLabel("Click on the button above to start this process.", self), font=QFont("Helvetica", 10)), alignment=Qt.AlignmentFlag.AlignCenter)
    layout.addWidget(util.apply_style(QLabel("(if your video is light enough you can also send it to info@zebrazoom.org without reducing its size)", self), font=QFont("Helvetica", 10)), alignment=Qt.AlignmentFlag.AlignCenter)

    startPageBtn = util.apply_style(QPushButton("Go to the start page", self), background_color=util.LIGHT_CYAN)
    startPageBtn.clicked.connect(lambda: controller.show_frame("StartPage"))
    layout.addWidget(startPageBtn, alignment=Qt.AlignmentFlag.AlignCenter)

    self.setLayout(layout)


class VideoToTroubleshootSplitVideo(QWidget):
  def __init__(self, controller):
    super().__init__(controller.window)
    self.controller = controller

    layout = QVBoxLayout()
    layout.addWidget(util.apply_style(QLabel("Ok, your sub-video has been saved in the folder you chose. You can now send that sub-video to info@zebrazoom.org", self), font=QFont("Helvetica", 10)), alignment=Qt.AlignmentFlag.AlignCenter)
    startPageBtn = util.apply_style(QPushButton("Go to the start page", self), background_color=util.LIGHT_CYAN)
    startPageBtn.clicked.connect(lambda: controller.show_frame("StartPage"))
    layout.addWidget(startPageBtn, alignment=Qt.AlignmentFlag.AlignCenter)

    self.setLayout(layout)



import numpy as np
from scipy.interpolate import UnivariateSpline

def getTailAngles(curbout, smoothingFactor, nbFramesTakenIntoAccount, numberOfBendsIncludedForMaxDetect):

  fin = curbout["BoutEnd"] - curbout["BoutStart"] + 1

  if type(curbout["Bend_Timing"]) == list and len(curbout["Bend_Timing"]) > 2:
    
    Bend_Timing = curbout["Bend_Timing"].copy()
    Bend_Timing.insert(0, 1)
    Bend_Timing.append(fin)
    Bend_Amplitude = curbout["Bend_Amplitude"].copy()
    Bend_Amplitude.insert(0, 0)
    Bend_Amplitude.append(0)
    TailAngle = curbout["TailAngle_smoothed"].copy()
    l = len(Bend_Timing) - 1
    
    if numberOfBendsIncludedForMaxDetect == -1 or numberOfBendsIncludedForMaxDetect >= len(Bend_Amplitude):
      argmax = np.argmax(np.abs(Bend_Amplitude))
    else:
      Bend_Amplitude_Array = Bend_Amplitude[0:numberOfBendsIncludedForMaxDetect]
      argmax               = np.argmax(np.abs(Bend_Amplitude_Array))
    
    maxAbsBendValue = Bend_Amplitude[argmax]
    tailAngle = 0
    if maxAbsBendValue < 0:
      tailAngle = (-np.array(curbout["TailAngle_smoothed"])).tolist()
    else:
      tailAngle = curbout["TailAngle_smoothed"]
  
  else:
    
    tailAngle = [0]*fin
  
  tailAngle3 = [0]*nbFramesTakenIntoAccount
  
  sizeTabNotZero = min(fin, nbFramesTakenIntoAccount)
  tailAngle3[0:sizeTabNotZero]  = tailAngle[0:sizeTabNotZero]

  return tailAngle3



import numpy as np
import math
import cvui
import json
import cv2
import os

def calculateAngle(xStart, yStart, xEnd, yEnd):
  vx = xEnd - xStart
  vy = yEnd - yStart
  if vx == 0:
    if vy > 0:
      lastFirstTheta = math.pi/2
    else:
      lastFirstTheta = (3*math.pi)/2
  else:
    lastFirstTheta = np.arctan(abs(vy/vx))
    if (vx < 0) and (vy >= 0):
      lastFirstTheta = math.pi - lastFirstTheta
    elif (vx < 0) and (vy <= 0):
      lastFirstTheta = lastFirstTheta + math.pi
    elif (vx > 0) and (vy <= 0):
      lastFirstTheta = 2*math.pi - lastFirstTheta
  return lastFirstTheta
  
def distBetweenThetas(theta1, theta2):
  diff = 0
  if theta1 > theta2:
    diff = theta1 - theta2
  else:
    diff = theta2 - theta1
  if diff > math.pi:
    diff = (2 * math.pi) - diff
  return diff

def assignValueIfBetweenRange(value, minn, maxx):
  if value < minn:
    return minn
  if value > maxx:
    return maxx
  return value

def addBlackLineToImgSetParameters(hyperparameters, frame, videoName):
  
  hyperparametersToSave = {"addBlackLineToImg_Width": 0}
  
  frame2 = frame.copy()
  # frame2 = 255 - frame2
  quartileChose = hyperparameters["outputValidationVideoContrastImprovementQuartile"]
  lowVal  = int(np.quantile(frame2, quartileChose))
  highVal = int(np.quantile(frame2, 1 - quartileChose))
  frame2[frame2 < lowVal]  = lowVal
  frame2[frame2 > highVal] = highVal
  frame2 = frame2 - lowVal
  mult  = np.max(frame2)
  frame2 = frame2 * (255/mult)
  frame2 = frame2.astype('uint8')
  
  addNewSegment  = True
  
  while addNewSegment:
    
    WINDOW_NAME = "Click on beginning of segment to set to black pixels"
    cvui.init(WINDOW_NAME)
    cv2.moveWindow(WINDOW_NAME, 0, 0)
    cvui.imshow(WINDOW_NAME, frame2)
    plus = 1
    while not(cvui.mouse(WINDOW_NAME, cvui.CLICK)):
      cursor = cvui.mouse(WINDOW_NAME)
      if cv2.waitKey(20) != -1:
        cvui.imshow(WINDOW_NAME, frame2)
        plus = plus + 1
    cv2.destroyWindow(WINDOW_NAME)
    startPoint = [cursor.x, cursor.y]
    
    WINDOW_NAME = "Click on the end of the segment to set to black pixels"
    cvui.init(WINDOW_NAME)
    cv2.moveWindow(WINDOW_NAME, 0, 0)
    cvui.imshow(WINDOW_NAME, frame2)
    plus = 1
    while not(cvui.mouse(WINDOW_NAME, cvui.CLICK)):
      cursor = cvui.mouse(WINDOW_NAME)
      frame3 = frame2.copy()
      frame3 = cv2.line(frame3, (startPoint[0], startPoint[1]), (cursor.x, cursor.y), (255, 255, 255), hyperparameters["addBlackLineToImg_Width"])
      cvui.imshow(WINDOW_NAME, frame3)
      plus = plus + 1
      cv2.waitKey(20)
      del frame3
    cv2.destroyWindow(WINDOW_NAME)
    endPoint = [cursor.x, cursor.y]  
    
    if not("imagePreProcessMethod" in hyperparameters) or type(hyperparameters["imagePreProcessMethod"]) == int or len(hyperparameters["imagePreProcessMethod"]) == 0:
      hyperparameters["imagePreProcessMethod"]     = ["setImageLineToBlack"]
      hyperparameters["imagePreProcessParameters"] = [[startPoint[0], startPoint[1], endPoint[0], endPoint[1], hyperparameters["addBlackLineToImg_Width"]]]
      hyperparametersToSave["imagePreProcessMethod"]     = ["setImageLineToBlack"]
      hyperparametersToSave["imagePreProcessParameters"] = [[startPoint[0], startPoint[1], endPoint[0], endPoint[1], hyperparameters["addBlackLineToImg_Width"]]]
    else:
      hyperparameters["imagePreProcessMethod"] = hyperparameters["imagePreProcessMethod"] + ["setImageLineToBlack"]
      hyperparameters["imagePreProcessParameters"] += [[startPoint[0], startPoint[1], endPoint[0], endPoint[1], hyperparameters["addBlackLineToImg_Width"]]]
      hyperparametersToSave["imagePreProcessMethod"]     += ["setImageLineToBlack"]
      hyperparametersToSave["imagePreProcessParameters"] += [[startPoint[0], startPoint[1], endPoint[0], endPoint[1], hyperparameters["addBlackLineToImg_Width"]]]
    
    print('addBlackLineToImg_Width is at', hyperparameters["addBlackLineToImg_Width"],', so in the configuration file, "imagePreProcessMethod" has just been set to ["setImageLineToBlack"] and "imagePreProcessParameters" has just been set to', str(hyperparameters["imagePreProcessParameters"]))
    
    WINDOW_NAME = "Do you want to add another segment?"
    cvui.init(WINDOW_NAME)
    cv2.moveWindow(WINDOW_NAME, 0, 0)
    
    frame2 = cv2.line(frame2, (startPoint[0], startPoint[1]), (endPoint[0], endPoint[1]), (255, 255, 255), hyperparameters["addBlackLineToImg_Width"])
    frame3 = frame2.copy()
    
    answerYes = False
    answerNo  = False
    
    while not(answerYes) and not(answerNo):
      answerYes = cvui.button(frame3, 10, 10, "I want to add another segment.")
      answerNo  = cvui.button(frame3, 10, 40, "I've added enough segments.")
      cvui.imshow(WINDOW_NAME, frame3)
      cv2.waitKey(20)
    cv2.destroyAllWindows()
    del frame3
    
    addNewSegment = answerYes
    
  
  with open(os.path.join(os.path.join(hyperparameters["outputFolder"], videoName), 'parametersToAddToConfigFileForBlackLine.json'), 'w') as outfile:
    json.dump(hyperparametersToSave, outfile)
  
  return hyperparameters



import os
import scipy.io
import pandas as pd
import json
import numpy as np
np.warnings.filterwarnings('ignore', category=np.VisibleDeprecationWarning)
from zebrazoom.dataAnalysis.datasetcreation.getDynamicParameters import getDynamicParameters
from zebrazoom.dataAnalysis.datasetcreation.getTailAngles import getTailAngles
from zebrazoom.dataAnalysis.datasetcreation.getInstaSpeed import getInstaSpeed
from zebrazoom.dataAnalysis.datasetcreation.getInstaHeadingDiff import getInstaHeadingDiff
from zebrazoom.dataAnalysis.datasetcreation.getInstaHorizontalDisplacement import getInstaHorizontalDisplacement
from zebrazoom.dataAnalysis.datasetcreation.getGlobalParameters import getGlobalParameters
from zebrazoom.dataAnalysis.datasetcreation.getTailLength  import getTailLength
from zebrazoom.dataAnalysis.datasetcreation.getTailLength2 import getTailLength2
from zebrazoom.dataAnalysis.datasetcreation.getTailAngleRecalculated import getTailAngleRecalculated
from zebrazoom.dataAnalysis.datasetcreation.getTailAngleRecalculated2 import getTailAngleRecalculated2
from zebrazoom.dataAnalysis.datasetcreation.gatherInitialRawData import gatherInitialRawData
import pickle

def createDataFramePerFrame(dataframeOptions):

  # Gathering user-inputed information about how to create the dataframe of parameters for the whole set of videos
  
  pathToExcelFile                   = dataframeOptions['pathToExcelFile']
  nameOfFile                        = dataframeOptions['nameOfFile']
  fileExtension                     = dataframeOptions['fileExtension']
  smoothingFactor                   = dataframeOptions['smoothingFactorDynaParam']
  resFolder                         = dataframeOptions['resFolder']
  numberOfBendsIncludedForMaxDetect = dataframeOptions['numberOfBendsIncludedForMaxDetect']
  minNbBendForBoutDetect            = dataframeOptions['minNbBendForBoutDetect']
  computeTailAngleParamForCluster   = dataframeOptions['computeTailAngleParamForCluster']
  computeMassCenterParamForCluster  = dataframeOptions['computeMassCenterParamForCluster']
  
  defaultZZoutputFolderPath = dataframeOptions['defaultZZoutputFolderPath'] if 'defaultZZoutputFolderPath' in dataframeOptions else ''
  
  computetailAnglesRecalculatedParamsForCluster = dataframeOptions["computetailAnglesRecalculatedParamsForCluster"] if 'computetailAnglesRecalculatedParamsForCluster' in dataframeOptions else False
  
  keepSpeedDistDurWhenLowNbBends = int(dataframeOptions['keepSpeedDistDurWhenLowNbBends']) if 'keepSpeedDistDurWhenLowNbBends' in dataframeOptions else 1
  
  frameStepForDistanceCalculation = int(dataframeOptions['frameStepForDistanceCalculation']) if ('frameStepForDistanceCalculation' in dataframeOptions) and len(dataframeOptions['frameStepForDistanceCalculation']) else 4
  
  tailAngleKinematicParameterCalculation = int(dataframeOptions['tailAngleKinematicParameterCalculation']) if 'tailAngleKinematicParameterCalculation' in dataframeOptions else 0
  
  saveRawDataInAllBoutsSuperStructure = int(dataframeOptions['saveRawDataInAllBoutsSuperStructure']) if 'saveRawDataInAllBoutsSuperStructure' in dataframeOptions else 0
  
  saveAllBoutsSuperStructuresInMatlabFormat = int(dataframeOptions['saveAllBoutsSuperStructuresInMatlabFormat']) if 'saveAllBoutsSuperStructuresInMatlabFormat' in dataframeOptions else 0
  
  getTailAngleSignMultNormalized = int(dataframeOptions['getTailAngleSignMultNormalized']) if 'getTailAngleSignMultNormalized' in dataframeOptions else 0
  
  
  excelFile = pd.read_excel(os.path.join(pathToExcelFile, nameOfFile + fileExtension))
  
  # Creating labels of columns of dataframe
  # General basic information
  basicInformation = ['numFrame', 'Trial_ID', 'Well_ID', 'NumBout', 'BoutStart', 'BoutEnd', 'Condition', 'Genotype']
  # Tail angle related parameters for clustering
  instaTBF   = ['instaTBF']
  instaAmp   = ['instaAmp']
  instaAsym  = ['instaAsym']
  tailAngles = ['tailAngles']
  # Center of mass related parameters for clustering
  instaSpeed       = ['instaSpeed']
  instaHeadingDiff = ['instaHeadingDiff']
  instaHorizDispl  = ['instaHorizDispl']
  # Assembling columns
  dfCols = basicInformation + tailAngles

  if computeTailAngleParamForCluster:
    dfCols = dfCols + instaTBF + instaAmp + instaAsym
  if computeMassCenterParamForCluster:
    dfCols = dfCols + instaSpeed + instaHeadingDiff + instaHorizDispl

  if computeTailAngleParamForCluster or computeMassCenterParamForCluster or computetailAnglesRecalculatedParamsForCluster:
    dfCols = dfCols + ['classification']
  numberOfParameters = len(dfCols)
  
  # Creating an empty dataframe then filling it with the parameters for the whole set of videos
  print("Calculating and storing all parameters:")
  dfParam = pd.DataFrame(columns=dfCols)
  genotypes  = []
  conditions = []
  # Going through each video listed in the excel file
  for videoId in range(0, len(excelFile)):
    if excelFile.loc[videoId, 'path'] == "defaultZZoutputFolder":
      path    = os.path.join(defaultZZoutputFolderPath, excelFile.loc[videoId, 'trial_id'])
    else:
      path    = os.path.join(excelFile.loc[videoId, 'path'], excelFile.loc[videoId, 'trial_id'])
    trial_id  = excelFile.loc[videoId, 'trial_id']
    fq        = excelFile.loc[videoId, 'fq']
    pixelsize = excelFile.loc[videoId, 'pixelsize']
    condition = excelFile.loc[videoId, 'condition']
    condition = eval('[' + condition + ']')
    condition = condition[0]
    genotype  = excelFile.loc[videoId, 'genotype']
    genotype = eval('[' + genotype + ']')
    genotype = genotype[0]
    include   = excelFile.loc[videoId, 'include']
    include = eval('[' + include + ']')
    include = include[0]
    
    with open(os.path.join(path, 'results_' + trial_id + '.txt')) as f:
      supstruct = json.load(f)
    
    # Going through each well of the video
    for Well_ID, Cond in enumerate(condition):
      if include[Well_ID]:
        print("trial_id:", trial_id, " ; Well_ID:", Well_ID)
        dfParamForWell = pd.DataFrame(columns=dfCols)
        curBoutId = 0
        # Going through each animal present in the well
        for fishId in range(0, len(supstruct["wellPoissMouv"][Well_ID])):
          # Going through each bout performed by the animal
          for NumBout, dataForBout in enumerate(supstruct["wellPoissMouv"][Well_ID][fishId]):
            if not("flag" in dataForBout) or dataForBout["flag"] == 0:
              # Calculating specified parameters for that bout
              if "Bend_Timing" in dataForBout and type(dataForBout["Bend_Timing"]) == list and len(dataForBout["Bend_Timing"]) >= minNbBendForBoutDetect:
                
                # Initial basic information
                
                toPutInDataFrameColumn = basicInformation
                
                if not(genotype[Well_ID] in genotypes):
                  genotypes.append(genotype[Well_ID])
                if not(condition[Well_ID] in conditions):
                  conditions.append(condition[Well_ID])
                
                nbFramesTakenIntoAccount = dataForBout['BoutEnd'] - dataForBout['BoutStart']
                
                # Tail angles
                
                if getTailAngleSignMultNormalized:
                  
                  toPutInDataFrameColumn = toPutInDataFrameColumn + tailAngles
                  
                  tailAnglesData         = getTailAngles(dataForBout, smoothingFactor, nbFramesTakenIntoAccount, numberOfBendsIncludedForMaxDetect)
                
                # Calculate "dynamic" tail angle related parameters for clustering
                
                if computeTailAngleParamForCluster:
                
                  toPutInDataFrameColumn = toPutInDataFrameColumn + instaTBF + instaAmp + instaAsym
                  
                  tailAngleParamForCluster = getDynamicParameters(dataForBout, smoothingFactor, nbFramesTakenIntoAccount, numberOfBendsIncludedForMaxDetect)
                
                # Calculate "dynamic" center of mass related parameters for clustering
                
                if computeMassCenterParamForCluster:
                
                  toPutInDataFrameColumn = toPutInDataFrameColumn + instaSpeed + instaHeadingDiff + instaHorizDispl
                  
                  instaSpeedVal       = getInstaSpeed(dataForBout, nbFramesTakenIntoAccount)
                  instaHeadingDiffVal = getInstaHeadingDiff(dataForBout, nbFramesTakenIntoAccount)
                  instaHorizDisplVal  = getInstaHorizontalDisplacement(dataForBout, nbFramesTakenIntoAccount)
                
                # Adding bout parameters to the dataframe created for the current well
                
                for idx in range(nbFramesTakenIntoAccount):
                  if computeMassCenterParamForCluster:
                    dfParamForWell.loc[curBoutId, toPutInDataFrameColumn] = [dataForBout['BoutStart']+idx, trial_id, Well_ID, NumBout, dataForBout['BoutStart'], dataForBout['BoutEnd'], condition[Well_ID], genotype[Well_ID], tailAnglesData[idx], tailAngleParamForCluster[idx], tailAngleParamForCluster[nbFramesTakenIntoAccount+idx], tailAngleParamForCluster[2*nbFramesTakenIntoAccount+idx], instaSpeedVal[idx], instaHeadingDiffVal[idx], instaHorizDisplVal[idx]]
                  else:
                    dfParamForWell.loc[curBoutId, toPutInDataFrameColumn] = [dataForBout['BoutStart']+idx, trial_id, Well_ID, NumBout, dataForBout['BoutStart'], dataForBout['BoutEnd'], condition[Well_ID], genotype[Well_ID], tailAnglesData[idx], tailAngleParamForCluster[idx], tailAngleParamForCluster[nbFramesTakenIntoAccount+idx], tailAngleParamForCluster[2*nbFramesTakenIntoAccount+idx]]
                  curBoutId = curBoutId + 1
        
        # Adding dataframe created for the current frame to the dataframe for the whole set of videos
        dfParam = pd.concat([dfParam, dfParamForWell])
  
  # Saving the dataframe
  dfParam = dfParam.reset_index()
  
  # Saving dataframe for the whole set of videos as a pickle file
  outfile = open(os.path.join(resFolder, nameOfFile), 'wb')
  pickle.dump(dfParam,outfile)
  outfile.close()
  
  return [conditions, genotypes, nbFramesTakenIntoAccount]
